"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.output_tools import *
import numpy as np
from os import sep


class TestOutput(TestCase):
    """ Test how well we parse the outputs """
    parser_objects = []
    parser_objects.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_1.hdf5",
                                       flim=True))
    parser_objects.append(OutputParser(output_dict={'memory': True}, filename="tests_data" + sep + f"output_2.hdf5",
                                       reprate=160.3e6, flim=True, z_pixels=10, num_of_frames=2))
    parser_objects.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_1.hdf5",
                                       num_of_frames=2))

    def test_bins_bet_pulses(self):
        bins_bet_pulses = [16, 8, 1]
        for obj, bins in zip(self.parser_objects, bins_bet_pulses):
            self.assertEqual(obj.bins_bet_pulses, bins)


    def test_determine_shape(self):
        shapes = [(1, 512, 512, 16), (2, 512, 512, 10, 8), (2, 512, 512)]
        for shape, obj in zip(shapes, self.parser_objects):
            self.assertEqual(shape, obj.determine_data_shape_full())



"""
__author__ = Hagai Hargil
"""
from unittest import TestCase
from pysight.rectify_lines import LineRectifier
import numpy as np


class TestLineRectifier(TestCase):

    def test_normal_line_signal_from_zero_unidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        raw_lines = np.arange(start=0, stop=100, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_normal_line_signal_from_zero_bidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        raw_lines = np.arange(start=0, stop=100, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify())[:-1], list(raw_lines)[:-1])

    def test_normal_line_signal_no_zero_unidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        raw_lines = np.arange(start=10, stop=100, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_normal_line_signal_no_zero_bidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        raw_lines = np.arange(start=10, stop=100, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_normal_line_signal_with_noise_unidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        step = 100
        raw_lines = np.arange(start=500, stop=15000, step=step, dtype=np.uint64)
        raw_lines += ((np.random.rand(raw_lines.shape[0]) - 0.5) * step/25).astype(np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=False)

        for idx, (val1, val2) in enumerate(zip(rect.rectify()[:-1], raw_lines[:-1])):
            self.assertGreaterEqual(1, np.abs(float(val1)-float(val2)))

    def test_normal_line_signal_with_noise_bidir(self):
        """
        Non corrupt signal to test
        :return:
        """
        step = 100
        raw_lines = np.arange(start=500, stop=15000, step=step, dtype=np.uint64)
        raw_lines += ((np.random.rand(raw_lines.shape[0]) - 0.5) * step/25).astype(np.uint64)
        num_lines = len(raw_lines[:-1])
        rect = LineRectifier(lines=raw_lines[:-1], x_pixels=num_lines,
                             bidir=True)

        for idx, (val1, val2) in enumerate(zip(rect.rectify()[:-1], raw_lines[:-1])):
            self.assertGreaterEqual(1, np.abs(float(val1)-float(val2)))

    def test_single_missing_line_middle_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, 2)
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_single_missing_line_middle_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, 2)
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_line_middle_single_instance_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 3, 4])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_line_middle_single_instance_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 3, 4])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_single_lines_middle_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 10, 100, 200, 300, 234])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_single_lines_middle_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 10, 100, 200, 300, 234])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_lines_middle_unidir(self):
        raw_lines = np.arange(start=0, stop=10000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 10, 100, 101, 102, 200, 201, 300, 234])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_multiple_lines_middle_bidir(self):
        raw_lines = np.arange(start=0, stop=10000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [2, 10, 100, 101, 102, 200, 201, 300, 234])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_line_from_end_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_line_from_end_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_lines_from_end_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2, len(raw_lines)-3, len(raw_lines)-4])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_lines_from_end_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2, len(raw_lines)-3, len(raw_lines)-4])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_lines_from_end_and_middle_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2, len(raw_lines)-3, len(raw_lines)-4,
                                          10, 20, 21, 22, 51])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_lines_from_end_and_middle_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [len(raw_lines)-2, len(raw_lines)-3, len(raw_lines)-4,
                                          10, 20, 21, 22, 51])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_issue_with_first_diff_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [1, 2])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_issue_with_first_diff_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [1, 2])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_issue_with_first_diff_and_more_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [1, 2, 10, 40, 41, 42, 100, 101])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_issue_with_first_diff_and_more_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [1, 2, 10, 40, 41, 42, 100, 101])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_starting_lines_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [0, 1, 2, 3])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_starting_lines_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [0, 1, 2, 3])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_head_tail_unidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [0, 1, 2, 3, len(raw_lines)-2, len(raw_lines)-3])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        expected_lines = np.array([0, 3, 6, 10, 13, 16], dtype=np.uint64)
        expected_lines = np.concatenate((expected_lines, raw_lines[4:-2]))
        self.assertSequenceEqual(list(rect.rectify()), list(expected_lines))

    def test_missing_head_tail_bidir(self):
        raw_lines = np.arange(start=0, stop=1000, step=5, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [0, 1, 2, 3, len(raw_lines)-2, len(raw_lines)-3])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        expected_lines = np.array([0, 3, 6, 10, 13, 16], dtype=np.uint64)
        expected_lines = np.concatenate((expected_lines, raw_lines[4:-2]))
        self.assertSequenceEqual(list(rect.rectify()), list(expected_lines))

    def test_missing_with_one_line_ok_unidir(self):
        raw_lines = np.arange(start=0, stop=2000, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 5])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_one_line_ok_bidir(self):
        raw_lines = np.arange(start=0, stop=2000, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 5])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_ok_unidir(self):
        raw_lines = np.arange(start=0, stop=2000, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_ok_bidir(self):
        raw_lines = np.arange(start=0, stop=2000, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_and_one_ok_unidir(self):
        raw_lines = np.arange(start=0, stop=2500, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6, 9, 11])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_and_one_ok_bidir(self):
        raw_lines = np.arange(start=0, stop=2500, step=10, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6, 9, 11])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_or_one_ok_unidir(self):
        raw_lines = np.arange(start=0, stop=2500, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6, 11])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=False)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))

    def test_missing_with_two_lines_or_one_ok_bidir(self):
        raw_lines = np.arange(start=0, stop=2500, step=50, dtype=np.uint64)
        num_lines = len(raw_lines[:-1])
        del_lines = np.delete(raw_lines, [3, 6, 11])
        rect = LineRectifier(lines=del_lines[:-1], x_pixels=num_lines,
                             bidir=True)
        self.assertSequenceEqual(list(rect.rectify()), list(raw_lines))



"""
__author__ = Hagai Hargil
"""
import pandas as pd
import numpy as np
import attr
from typing import Dict, Tuple, List
from numba import jit, int64, uint64
from pysight.validation_tools import  rectify_photons_in_uneven_lines
from pysight.tag_tools_v2 import TagPipeline
from attr.validators import instance_of
from scipy.optimize import curve_fit
from scipy.signal import find_peaks_cwt


@attr.s(slots=True)
class Allocate(object):
    """
    Create pipeline of analysis of lst files.
    """
    # TODO: Variable documentation
    dict_of_inputs     = attr.ib(validator=instance_of(dict))
    df_photons         = attr.ib(validator=instance_of(pd.DataFrame))
    num_of_channels    = attr.ib(default=1, validator=instance_of(int))
    laser_freq         = attr.ib(default=80.3e6, validator=instance_of(float))
    binwidth           = attr.ib(default=800e-12, validator=instance_of(float))
    bidir              = attr.ib(default=False, validator=instance_of(bool))
    tag_freq           = attr.ib(default=189e3, validator=instance_of(float))
    tag_pulses         = attr.ib(default=1, validator=instance_of(int))
    phase              = attr.ib(default=-2.78, validator=instance_of(float))
    keep_unidir        = attr.ib(default=False, validator=instance_of(bool))
    flim               = attr.ib(default=False, validator=instance_of(bool))
    exp_params         = attr.ib(default={}, validator=instance_of(dict))
    censor             = attr.ib(default=False, validator=instance_of(bool))
    dict_of_data       = attr.ib(default={}, validator=instance_of(dict))
    tag_interp_ok      = attr.ib(default=False, validator=instance_of(bool))
    tag_to_phase       = attr.ib(default=True, validator=instance_of(bool))
    tag_offset         = attr.ib(default=0, validator=instance_of(int))
    sorted_indices     = attr.ib(init=False)

    def run(self):
        """ Pipeline of analysis """
        print('Channels of events found. Allocating photons to their frames and lines...')
        # Unidirectional scan - create fake lines
        if self.bidir:
            pass
        else:
            self.__add_unidirectional_lines()
        self.allocate_photons()
        print('Relative times calculated. Creating Movie object...')
        # Censor correction addition:
        if 'Laser' not in self.dict_of_data.keys():
            self.dict_of_data['Laser'] = 0
            # if self.use_sweeps:
            #     out = self.train_dataset()

    def allocate_photons(self):
        """
        Returns a dataframe in which each photon is a part of a frame, line and possibly laser pulse
        """

        # Preparations
        irrelevant_keys = {'PMT1', 'PMT2', 'TAG Lens'}
        relevant_keys = set(self.dict_of_data.keys()) - irrelevant_keys
        column_heads = {'Lines': 'time_rel_line_pre_drop', 'Frames': 'time_rel_frames', 'Laser': 'time_rel_pulse'}

        # Main loop - Sort lines and frames for all photons and calculate relative time
        for key in reversed(sorted(relevant_keys)):
            sorted_indices = numba_search_sorted(self.dict_of_data[key].loc[:, 'abs_time'].values,
                                                 self.df_photons.loc[:, 'abs_time'].values)
            self.df_photons[key] = self.dict_of_data[key].iloc[sorted_indices, 0].values  # columns 0 is abs_time,
            # but this .iloc method is amazingly faster than .loc
            positive_mask = sorted_indices >= 0
            # drop photons that came before the first line
            self.df_photons = self.df_photons.iloc[positive_mask].copy()
            # relative time of each photon in accordance to the line\frame\laser pulse
            self.df_photons[column_heads[key]] = self.df_photons['abs_time'] - self.df_photons[key]
            self.sorted_indices = sorted_indices[sorted_indices >= 0]
            if 'Lines' == key:
                self.df_photons = rectify_photons_in_uneven_lines(df=self.df_photons,
                                                                  sorted_indices=self.sorted_indices,
                                                                  lines=self.dict_of_data['Lines'].loc[:, 'abs_time'],
                                                                  bidir=self.bidir,
                                                                  keep_unidir=self.keep_unidir)

            if 'Laser' != key:
                self.df_photons.loc[:, key] = self.df_photons.loc[:, key].astype('category')
            self.df_photons.set_index(keys=key, inplace=True, append=True, drop=True)

        assert len(self.df_photons) > 0
        assert np.all(self.df_photons.iloc[:, 0].values >= 0)  # finds NaNs as well

        # Deal with TAG lens interpolation
        try:
            tag = self.dict_of_data['TAG Lens'].loc[:, 'abs_time']
        except KeyError:
            pass
        else:
            print('Interpolating TAG lens data...')
            tag_pipe = TagPipeline(photons=self.df_photons, tag_pulses=tag, freq=self.tag_freq,
                                   binwidth=self.binwidth, num_of_pulses=self.tag_pulses,
                                   to_phase=self.tag_to_phase, offset=self.tag_offset)
            tag_pipe.run()
            self.df_photons = tag_pipe.photons
            self.tag_interp_ok = tag_pipe.finished_pipe

            print('TAG lens interpolation finished.')

        # Deal with laser pulses interpolation
        if self.flim:
            self.df_photons, rel_time = self.__interpolate_laser(self.df_photons)
            if self.censor:
                self.exp_params = self.__fit_data_to_exponent(rel_time)

    def __nano_flim_exp(self, x, a, b, c):
        """ Exponential function for FLIM and censor correction """
        return a * np.exp(-b * x) + c

    def __fit_data_to_exponent(self, list_of_channels: List) -> Dict:
        """
        Take the data after modulu BINS_BETWEEN_PULSES, in each channel,
         and fit an exponential decay to it, with some lifetime.
        :return: (A, b, C): Parameters of the fit A * exp( -b * x ) + C as a numpy array,
        inside a dictionary with the channel number as its key.
        """
        params = {}
        for chan, data_of_channel in enumerate(list_of_channels, 1):
            yn = np.histogram(data_of_channel, 16)[0]
            peakind = find_peaks_cwt(yn, np.arange(1, 10))
            if self.__requires_censoring(data=yn[peakind[0]:]):
                min_value = min(yn[yn > 0])
                max_value = yn[peakind[0]]
                y_filt = yn[peakind[0]:]
                x = np.arange(len(y_filt))
                popt, pcov = curve_fit(self.__nano_flim_exp, x, y_filt, p0=(max_value, 1 / 3.5, min_value),
                                       maxfev=10000)
                params[chan] = popt
            else:
                continue

        return params

    def __requires_censoring(self, data: np.ndarray):
        """
        Method to determine if we should undergo the censor correction process
        :param data: Bins of histogram from their peak onward.
        :return: boolean value
        """
        diffs = np.diff(data)
        if len(diffs) == 0:
            return True
        if np.all(diffs <= 0):  # No censoring occurred
            return False
        else:
            first_idx = np.argwhere(diffs >= 0)[0][0]
            diffs2 = diffs[first_idx:]
            if np.all(diffs2 >= 0) or np.all(diffs2 <= 0):
                return True
            else:  # either false alarm, or a third photon is on its way
                self.__requires_censoring(diffs2)
                return True

    def __add_unidirectional_lines(self):
        """
        For unidirectional scans fake line signals have to be inserted for us to identify forward- and
        back-phase photons.
        """

        length_of_lines = self.dict_of_data['Lines'].shape[0]
        new_line_arr = np.zeros(length_of_lines * 2 - 1)
        new_line_arr[::2] = self.dict_of_data['Lines'].loc[:, 'abs_time'].values
        new_line_arr[1::2] = self.dict_of_data['Lines'].loc[:, 'abs_time'] \
                                 .rolling(window=2).mean()[1:]

        self.dict_of_data['Lines'] = pd.DataFrame(new_line_arr, columns=['abs_time'],
                                                  dtype='uint64')

    def __interpolate_laser(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, List]:
        """
        Assign a time relative to a laser pulse for each photon.
        :param df: Dataframe with data for each photon.
        Assuming that the clock is synced to a 10 MHz signal.
        :return: Modified dataframe.
        """
        TEN_MEGAHERTZ_IN_BINS = 251
        rel_time = []
        for chan in range(1, self.num_of_channels + 1):
            rel_time251 = df.xs(key=chan, level='Channel', drop_level=False)[
                              'abs_time'].values % TEN_MEGAHERTZ_IN_BINS
            rel_time_per_pulse = rel_time251 % np.ceil(1 / (self.binwidth * self.laser_freq))
            rel_time.append(np.uint8(rel_time_per_pulse))
            df.loc[chan, 'time_rel_pulse'] = np.uint8(rel_time_per_pulse)

        return df, rel_time

    def __add_phase_offset_to_bidir_lines(self):
        """
        Uneven lines in a bidirectional scanning mode have to be offsetted.
        """
        phase_in_seconds = self.phase * 1e-6
        self.dict_of_data['Lines'].abs_time.iloc[1::2] -= np.uint64(phase_in_seconds/self.binwidth)

@jit(nopython=True, cache=True)
def numba_sorted(arr: np.array) -> np.array:
    """
    Sort an array with Numba. CURRENTLY NOT WORKING
    """

    arr.sort()
    return arr.astype(np.uint64)

@jit((int64[:](uint64[:], uint64[:])), nopython=True, cache=True)
def numba_search_sorted(input_sorted, input_values):
    """ Numba-powered searchsorted function. """
    return np.searchsorted(input_sorted, input_values) - 1



"""
__author__ = Hagai Hargil
"""
import logging
import os


def basic_logging(logger_name: str='PySightLogger'):
    """
    Create a basic logger with different handles.
    :param logger_name: Name of the logger
    :param filename: Filename of data
    :return: Logger object
    """
    # set up logging to file - see previous section for more details
    os.mkdir("./logs/")
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M',
                        filename=f"./logs/{logger_name}.log")
    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger('').addHandler(console)



"""
__author__ = Hagai Hargil
"""
from bokeh.plotting import output_file, show, figure
from bokeh.models import Div
from bokeh.layouts import gridplot, column
import numpy as np

def f(movie):
    hist_p, _ = np.histogram(np.ravel(movie.stack[1][9]), bins=range(15))
    hist_p_summed, _ = np.histogram(np.ravel(movie.summed_mem[1]), bins=range(15))

    one_frame = figure(title="Single Frame")
    one_frame.image(image=[movie.stack[1][9]], x=None, y=None, dw=[1], dh=[1])

    output_file(r"C:\Users\Hagai\to_ori_29-8-17.html", title="Histograms of photons")

    summed_frames = figure(title="Sum of Ten Frames")
    summed_frames.image(image=[movie.summed_mem[1]], x=None, y=None, dw=[1], dh=[1])

    hist_one = figure(y_axis_type="log")
    hist_one.vbar(x=range(14), width=0.5, bottom=1, top=hist_p)
    hist_one.yaxis.axis_label = "Number of pixels"
    hist_one.xaxis.axis_label = "Number of photons"

    hist_summed = figure(y_axis_type="log")
    hist_summed.vbar(x=range(14), width=0.5, bottom=1, top=hist_p_summed)
    hist_summed.xaxis.axis_label = "Number of photons"

    s = gridplot([[one_frame, summed_frames], [hist_one, hist_summed]])
    show(column(Div(text="<h2>Photon distributions, about 0.01 photons per laser pulse</h2>", width=1000),
                s))



"""
__author__ = Hagai Hargil
"""

import attr
from attr.validators import instance_of
from pysight.movie_tools import trunc_end_of_file
import numpy as np
import warnings
import h5py_cache
import os


@attr.s(slots=True)
class OutputParser(object):
    """
    Parse the wanted outputs and produce a dictionary with
    file pointers to the needed outputs
    """
    output_dict = attr.ib(validator=instance_of(dict))
    filename = attr.ib(validator=instance_of(str),
                       convert=trunc_end_of_file)
    num_of_frames = attr.ib(default=1, validator=instance_of(int))
    x_pixels = attr.ib(default=512, validator=instance_of(int))
    y_pixels = attr.ib(default=512, validator=instance_of(int))
    z_pixels = attr.ib(default=1, validator=instance_of(int))
    num_of_channels = attr.ib(default=1, validator=instance_of(int))
    flim = attr.ib(default=False, validator=instance_of(bool))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    reprate = attr.ib(default=80e6, validator=instance_of(float))
    lst_metadata = attr.ib(default={}, validator=instance_of(dict))
    file_pointer_created = attr.ib(default=True, validator=instance_of(bool))
    cache_size = attr.ib(default=10 * 1024**3, validator=instance_of(int))
    debug = attr.ib(default=False, validator=instance_of(bool))
    outputs = attr.ib(init=False)

    def run(self):
        """ Parse what the user required, creating a list of HDF5 dataset pointers for each channel """
        self.outputs = {}
        if not self.output_dict:
            return
        try:
            self.outputs['memory'] = self.output_dict['memory']
        except KeyError:
            pass
        f = self.__create_prelim_file()
        if f is not None:
            data_shape_full = self.determine_data_shape_full()
            self.__populate_hdf(f, data_shape_full=data_shape_full)

    def __create_prelim_file(self):
        """ Try to create a preliminary .hdf5 file. Cache improves IO performance """
        if 'stack' in self.output_dict or 'summed' in self.output_dict:
            try:
                split = os.path.splitext(self.filename)[0]
                debugged = '_DEBUG' if self.debug else ''
                fullfile = f'{split + debugged}.hdf5'
                f = h5py_cache.File(fullfile, 'w', chunk_cache_mem_size=self.cache_size, libver='latest', w0=1)
                self.outputs['filename'] = fullfile
            except PermissionError or OSError:
                self.file_pointer_created = False
                warnings.warn("Permission Error: Couldn't write data to disk.")
                return
            return f
        else:
            return

    def __populate_hdf(self, f, data_shape_full):
        """
        Generate files and add metadata to each group, write out the data in chunks
        f: File pointer
        """
        data_shape_summed = data_shape_full[1:]
        chunk_shape = list(data_shape_full)
        chunk_shape[0] = 1
        if 'stack' in self.output_dict:
            try:
                self.outputs['stack'] = [f.require_group('Full Stack')
                                          .require_dataset(name=f'Channel {channel}',
                                                           shape=data_shape_full,
                                                           dtype=np.uint8,
                                                           chunks=tuple(chunk_shape),
                                                           compression='gzip')
                                         for channel in range(1, self.num_of_channels + 1)]

                for key, val in self.lst_metadata.items():
                    for chan in range(self.num_of_channels):
                        self.outputs['stack'][chan].attrs.create(name=key, data=val.encode())

            except PermissionError or OSError:
                self.file_pointer_created = False
        if 'summed' in self.output_dict:
            try:
                self.outputs['summed'] = [f.require_group('Summed Stack')
                                           .require_dataset(name=f'Channel {channel}',
                                                            shape=data_shape_summed,
                                                            dtype=np.uint16,
                                                            chunks=True,
                                                            compression="gzip")
                                          for channel in range(1, self.num_of_channels + 1)]
                for key, val in self.lst_metadata.items():
                    for chan in range(self.num_of_channels):
                        self.outputs['summed'][chan].attrs.create(name=key, data=val.encode())

            except PermissionError or OSError:
                self.file_pointer_created = False

        f.close()
        if self.file_pointer_created is False:
            warnings.warn("Permission Error: Couldn't write data to disk.")

    @property
    def bins_bet_pulses(self) -> int:
        if self.flim:
            return int(np.ceil(1 / (self.reprate * self.binwidth)))
        else:
            return 1

    def determine_data_shape_full(self):
        """ Return the tuple that describes the shape of the final dataset """

        # Dimension order: [X, Y, Z, LIFETIME, FRAME]
        shape = np.squeeze(np.empty(shape=(self.num_of_frames,
                                          self.x_pixels,
                                          self.y_pixels,
                                          self.z_pixels,
                                          self.bins_bet_pulses),
                                   dtype=np.int8)).shape
        if self.num_of_frames == 1:  # Edge case in which the single frame is "squeezed out"
            return (1,) + shape
        return shape



"""
__author__ = Hagai Hargil
"""
import attr
from attr.validators import instance_of
import pandas as pd
import sys


@attr.s(slots=True)
class PhotonDF(object):
    """
    Create initial photon dataframe and set the channel as its index
    """
    dict_of_data    = attr.ib(validator=instance_of(dict))
    num_of_channels = attr.ib(default=1, validator=instance_of(int))

    def gen_df(self):
        """
        If a single PMT channel exists, create a df_photons object.
        Else, concatenate the two data channels into a single dataframe.
        :return pd.DataFrame: Photon data
        """

        assert 'Channel' in self.dict_of_data['PMT1'].columns

        try:
            df_photons = pd.concat([self.dict_of_data['PMT1'].copy(),
                                    self.dict_of_data['PMT2'].copy()], axis=0)
            self.num_of_channels = 2
        except KeyError:
            df_photons = self.dict_of_data['PMT1'].copy()
        except:
            print("Unknown error: ", sys.exc_info()[0])
        finally:
            df_photons.loc[:, 'Channel'] = df_photons.loc[:, 'Channel'].astype('category')
            df_photons.set_index(keys='Channel', inplace=True)

        return df_photons





"""
__author__ = Hagai Hargil
"""

import unittest
import numpy as np
import pandas as pd
from pysight.tag_tools_v2 import TagPeriodVerifier, TagPhaseAllocator, \
    TagPipeline, numba_digitize

class TestTagPipeline(unittest.TestCase):
    """
    Tests for TAG analysis functions
    """

    tag_data = pd.Series(np.arange(0, 200 * 6530, 6530))
    photons = pd.DataFrame([10, 100, 1000], columns=['abs_time'])
    def_pipe = TagPipeline(photons=photons, tag_pulses=tag_data)

    def test_preservation(self):
        photons = pd.DataFrame([-1, 10, 6531], columns=['abs_time'])
        pipe = TagPipeline(photons=photons, tag_pulses=self.tag_data)
        returned = pd.Series([0, 6530])
        self.assertSequenceEqual(returned.tolist(),
                                 pipe._TagPipeline__preserve_relevant_tag_pulses()
                                 .tolist())

    def test_preservation_without_zero(self):
        photons = pd.DataFrame([10, 6531], columns=['abs_time'])
        pipe = TagPipeline(photons=photons, tag_pulses=self.tag_data)
        returned = pd.Series([6530])
        self.assertSequenceEqual(returned.tolist(),
                                 pipe._TagPipeline__preserve_relevant_tag_pulses()
                                 .tolist())

class TestTagPeriodVerifier(unittest.TestCase):
    """ Test the Verifier class """

    tag_data = pd.Series(np.arange(0, 200 * 6530, 6530))
    freq = 189e3
    binwidth = 800e-12
    def_verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(200 * 6530))

    def test_bins_bet_pulses(self):
        self.assertEqual(6614, self.def_verifier.period)

    def test_allowed_noise(self):
        self.assertEqual(331, self.def_verifier.allowed_noise)

    def test_start_end_no_issues(self):
        tag_data = pd.Series(np.arange(0, 100, 10))
        freq = 0.1
        binwidth = 1.
        verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(100))
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = np.array([], dtype=np.int64)
        my_end = np.array([], dtype=np.int64)
        self.assertEqual(my_start.tolist(), ret_start.tolist())
        self.assertEqual(my_end.tolist(), ret_end.tolist())

    def test_start_end_no_zero(self):
        tag_data = pd.Series(np.arange(0, 300, 10))
        tag_data.drop([0, 5, 6], inplace=True)
        tag_data = tag_data.append(pd.Series([3, 9, 25]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.
        verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(300))
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = [0, 3, 6]
        my_end = [2, 5, 7]
        self.assertSequenceEqual(list(ret_start), my_start)
        self.assertSequenceEqual(list(ret_end), my_end)

    def test_start_end_adding_zero(self):
        tag_data = pd.Series(np.arange(5, 300, 10))
        tag_data.drop([1, 5, 7, 8], inplace=True)
        tag_data = tag_data.append(pd.Series([9, 27, 29, 31]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.
        verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(300))
        ret_start, ret_end = verifier._TagPeriodVerifier__obtain_start_end_idx()
        my_start = [0, 7]
        my_end = [6, 9]
        self.assertSequenceEqual(list(ret_start), my_start)
        self.assertSequenceEqual(list(ret_end), my_end)

    def test_fix_tag_pulses_adding_zero(self):
        tag_data = pd.Series(np.arange(0, 100, 10))
        tag_data.drop([0, 5, 6], inplace=True)
        tag_data = tag_data.append(pd.Series([3, 9, 25]))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.
        verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(100))
        my_start = [0, 3, 6]
        my_end = [2, 5, 7]
        verifier._TagPeriodVerifier__fix_tag_pulses(starts=my_start,
                                                    ends=my_end)
        self.assertSequenceEqual(list(verifier.tag.values),
                                 list(np.arange(0, 100, 10)))

    def test_fix_tag_pulses_no_zero_end_missing(self):
        tag_data = pd.Series(np.arange(5, 95, 10, dtype=np.uint64))
        tag_data.drop([1, 5, 7, 8], inplace=True)
        tag_data = tag_data.append(pd.Series([9, 27, 29, 31], dtype=np.uint64))
        tag_data = tag_data.sort_values().reset_index(drop=True)
        freq = 0.1
        binwidth = 1.
        verifier = TagPeriodVerifier(tag=tag_data, freq=freq, binwidth=binwidth,
                                     last_photon=np.uint64(85))
        my_start = [0, 7]
        my_end = [6, 8]
        verifier._TagPeriodVerifier__fix_tag_pulses(starts=my_start,
                                                    ends=my_end)
        self.assertSequenceEqual(list(verifier.tag.values),
                                 list(np.arange(5, 75, 10)))


class TestTagPhaseAllocator(unittest.TestCase):

    def test_tag_digitize(self):
        x = np.array([0.2, 6.4, 3.0, 1.6])
        bins = np.array([0.0, 1.0, 2.5, 4.0, 10.0])
        real_result = np.array([1, 4, 3, 2])
        result, _ = numba_digitize(x, bins)
        self.assertTrue(np.array_equal(real_result, result))

    def test_allocate_phase_1(self):
        photons = pd.DataFrame([0, 2*np.pi, 4*np.pi], columns=['abs_time'])
        tag = pd.Series([0, 2*np.pi, 4*np.pi, 6*np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        result = [1, 1, 1]
        for elem1, elem2 in zip(result, phaser.photons.Phase.tolist()):
            self.assertAlmostEqual(elem1, elem2, 6)

    def test_allocate_phase_2(self):
        photons = pd.DataFrame([np.pi, 3*np.pi, 5*np.pi], columns=['abs_time'])
        tag = pd.Series([0, 2*np.pi, 4*np.pi, 6*np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        for elem1, elem2 in zip([-1, -1, -1], phaser.photons.Phase.tolist()):
            self.assertAlmostEqual(elem1, elem2, 6)

    def test_allocate_phase_3(self):
        photons = pd.DataFrame([1, 2, 3], columns=['abs_time'])
        tag = pd.Series([0, 2*np.pi, 4*np.pi, 6*np.pi])
        phaser = TagPhaseAllocator(photons, tag)
        phaser.allocate_phase()
        normed_result = photons.abs_time / (2 * np.pi)
        true_result = np.sin(normed_result * 2 * np.pi + np.pi/2).astype(np.float32)
        for elem1, elem2 in zip(true_result.tolist(), phaser.photons.Phase.tolist()):
            self.assertAlmostEqual(elem1, elem2, 6)

if __name__ == '__main__':
        unittest.main()



"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
__author__ = Hagai Hargil
Created on Thu Oct 13 09:37:02 2016
"""


def main_data_readout(gui):
    """
    Main function that reads the lst file and processes its data.
    """
    from pysight.fileIO_tools import FileIO
    from pysight.tabulation_tools import Tabulate
    from pysight.allocation_tools import Allocate
    from pysight.movie_tools import Movie
    from pysight import timepatch_switch
    from pysight.output_tools import OutputParser
    from pysight.gating_tools import GatedDetection
    from pysight.photon_df_tools import PhotonDF
    from pysight.tag_bits_tools import ParseTAGBits
    import numpy as np

    # Read the file
    cur_file = FileIO(filename=gui.filename.get(), debug=gui.debug.get(), input_start=gui.input_start.get(),
                      input_stop1=gui.input_stop1.get(), input_stop2=gui.input_stop2.get(), binwidth=gui.binwidth.get(),
                      use_sweeps=gui.sweeps_as_lines.get())
    cur_file.run()

    # Create input structures
    dict_of_slices_hex = timepatch_switch.ChoiceManagerHex().process(cur_file.timepatch)
    # dict_of_slices_bin = timepatch_switch.ChoiceManagerBinary().process(cur_file.timepatch)  # Not supported

    # Process events into dataframe
    tabulated_data = Tabulate(timepatch=cur_file.timepatch, data_range=cur_file.data_range,
                              dict_of_inputs=cur_file.dict_of_input_channels, data=cur_file.data,
                              is_binary=cur_file.is_binary, num_of_frames=gui.num_of_frames.get(),
                              laser_freq=float(gui.reprate.get()), binwidth=float(gui.binwidth.get()),
                              dict_of_slices_hex=dict_of_slices_hex, dict_of_slices_bin=None,
                              use_tag_bits=gui.tag_bits.get(), use_sweeps=gui.sweeps_as_lines.get(),
                              time_after_sweep=cur_file.time_after, acq_delay=cur_file.acq_delay,
                              line_freq=gui.line_freq.get(), x_pixels=gui.x_pixels.get(),
                              y_pixels=gui.y_pixels.get(), bidir=gui.bidir.get(),
                              bidir_phase=gui.phase.get())
    tabulated_data.run()

    photon_df = PhotonDF(dict_of_data=tabulated_data.dict_of_data)
    tag_bit_parser = ParseTAGBits(dict_of_data=tabulated_data.dict_of_data, photons=photon_df.gen_df(),
                                  use_tag_bits=gui.tag_bits.get(), bits_dict=gui.tag_bits_dict)

    analyzed_struct = Allocate(dict_of_inputs=cur_file.dict_of_input_channels, bidir=gui.bidir.get(),
                               laser_freq=float(gui.reprate.get()), binwidth=float(gui.binwidth.get()),
                               tag_pulses=int(gui.tag_pulses.get()), phase=gui.phase.get(),
                               keep_unidir=gui.keep_unidir.get(), flim=gui.flim.get(),
                               censor=gui.censor.get(), dict_of_data=tabulated_data.dict_of_data,
                               df_photons=tag_bit_parser.gen_df(), num_of_channels=cur_file.num_of_channels,
                               tag_freq=float(gui.tag_freq.get()), tag_to_phase=True, tag_offset=gui.tag_offset.get())
    analyzed_struct.run()

    # Determine type and shape of wanted outputs, and open the file pointers there
    outputs = OutputParser(num_of_frames=len(np.unique(analyzed_struct.df_photons.index
                                                       .get_level_values('Frames')).astype(np.uint64)),
                           output_dict=gui.outputs, filename=gui.filename.get(),
                           x_pixels=gui.x_pixels.get(), y_pixels=gui.y_pixels.get(),
                           z_pixels=gui.z_pixels.get() if analyzed_struct.tag_interp_ok else 1,
                           num_of_channels=analyzed_struct.num_of_channels, flim=gui.flim.get(),
                           binwidth=gui.binwidth.get(), reprate=gui.reprate.get(),
                           lst_metadata=cur_file.lst_metadata, debug=gui.debug.get())
    outputs.run()

    if gui.gating.get():
        gated = GatedDetection(raw=analyzed_struct.df_photons, reprate=gui.reprate.get(),
                               binwidth=gui.binwidth.get())
        gated.run()

    data_for_movie = gated.data if gui.gating.get() else analyzed_struct.df_photons

    # Create a movie object
    final_movie = Movie(data=data_for_movie, x_pixels=int(gui.x_pixels.get()),
                        y_pixels=int(gui.y_pixels.get()), z_pixels=outputs.z_pixels,
                        reprate=float(gui.reprate.get()), name=gui.filename.get(),
                        binwidth=float(gui.binwidth.get()), bidir=gui.bidir.get(),
                        fill_frac=gui.fill_frac.get() if cur_file.fill_fraction == -1.0 else cur_file.fill_fraction,
                        outputs=outputs.outputs, censor=gui.censor.get(),
                        num_of_channels=analyzed_struct.num_of_channels, flim=gui.flim.get(),
                        lst_metadata=cur_file.lst_metadata, exp_params=analyzed_struct.exp_params,
                        line_delta=int(tabulated_data.line_delta), use_sweeps=gui.sweeps_as_lines.get(),
                        tag_as_phase=True, tag_freq=float(gui.tag_freq.get()), mirror_phase=gui.phase.get())

    final_movie.run()

    return analyzed_struct.df_photons, final_movie


def run():
    """
    Run the entire script.
    """
    from pysight.tkinter_gui_multiscaler import GUIApp
    from pysight.tkinter_gui_multiscaler import verify_gui_input

    gui = GUIApp()
    gui.root.mainloop()
    verify_gui_input(gui)
    return main_data_readout(gui)


def run_batch(foldername: str, glob_str: str="*.lst", recursive: bool=False):
    """
    Run PySight on all list files in the folder
    :param foldername: str - Main folder to run the analysis on.
    :param glob_str: String for the `glob` function to filter list files
    :param recursive: bool - Whether the search should be recursive.
    :return: None
    """

    import pathlib
    from pysight.tkinter_gui_multiscaler import GUIApp
    from pysight.tkinter_gui_multiscaler import verify_gui_input

    path = pathlib.Path(foldername)
    if not path.exists():
        raise UserWarning(f"Folder {foldername} doesn't exist.")
    if recursive:
        all_lst_files = path.rglob(glob_str)
        print(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            print(str(file))
        all_lst_files = path.rglob(glob_str)
    else:
        all_lst_files = path.glob(glob_str)
        print(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            print(str(file))
        all_lst_files = path.glob(glob_str)

    gui = GUIApp()
    gui.root.mainloop()
    gui.filename.set('.lst')  # no need to choose a list file
    verify_gui_input(gui)

    try:
        for lst_file in all_lst_files:
            gui.filename.set(str(lst_file))
            df, movie = main_data_readout(gui)
    except TypeError as e:
        print(e)


if __name__ == '__main__':
    df, movie = run()
    # run_batch(foldername=r"X:\Hagai", glob_str="*.lst", recursive=False)



import time
bins = self.bins_bet_pulses
self.num_of_vols = len(self.list_of_volume_times)-1
data_shape = list(np.squeeze(np.empty(shape=(self.num_of_vols,
                                             self.x_pixels,
                                             self.y_pixels,
                                             self.z_pixels,
                                             bins),
                                      dtype=np.int8)).shape)
data_shape[0] = 6
chunk_shape = list(data_shape)
chunk_shape[0] = 1
# file_loc = r"X:\Hagai\test.h5"
file_loc = r'C:\Users\Hagai\Documents\GitHub\python-pysight\test.h5'
with h5py_cache.File(file_loc, 'w', chunk_cache_mem_size=50 * 1024**2, w0=1,
                       libver='latest') as file:
    f = file.require_group('Full Stack').require_dataset(name=f'Channel {chan}',
                                                         shape=(data_shape),
                                                         dtype=np.uint8,
                                                         chunks=(1, 512, 512, 10, 16),
                                                         compression='gzip',
                                                         shuffle=False)
    self.stack[chan] = deque()
    for idx, vol in enumerate(self.gen_of_volumes(channel_num=chan)):
        print("Vol starting")
        start = time.time()
        hist, _ = vol.create_hist()
        print(f"Vol {idx} ending")
        self.stack[chan].append(hist)
        self.summed_mem[chan] += np.uint16(hist)
        if idx > data_shape[0]-2:
            break

    print("Stacking the stack...")
    start = time.time()
    self.stack[chan] = np.stack(self.stack[chan], axis=0)
    print(f"To create the stack with shape {self.stack[chan].shape} I took {time.time() - start} seconds.")
    print("Saving full stack to disk...")
    start = time.time()
    f[...] = self.stack[chan]
    print(f"Done after {time.time() - start} seconds.")



"""
__author__ = Hagai Hargil
"""
import numpy as np
import pandas as pd
import attr
from attr.validators import instance_of
import warnings
from enum import Enum, auto

class LineSignalType(Enum):
    VALID_UNIDIR = auto()
    VALID_BIDIR  = auto()
    EXTRA_LINES  = auto()
    SINGLE_PIXEL  = auto()


@attr.s(slots=True)
class LineRectifier:
    """
    Receives the input lines and makes sure that they're OK and will render a good volume
    Returns the np.array with the corrected lines.
    Errors - ValueError if signal can't be corrected.
             ValueError
    """
    lines             = attr.ib(validator=instance_of(np.ndarray))  # raw line signal
    x_pixels          = attr.ib(default=512, validator=instance_of(int))  # number of expected lines
    bidir             = attr.ib(default=False, validator=instance_of(bool))  # bidir or unidir scan
    ALLOWED_THRESHOLD = attr.ib(default=0.1, validator=instance_of(float))
    end_time          = attr.ib(default=np.uint64(1), validator=instance_of(np.uint64))
    rectified_lines   = attr.ib(init=False)  # returned signal
    __process_method  = attr.ib(init=False)

    def rectify(self) -> np.ndarray:
        """
        Main method to run the validation process on the line signal
        :return np.ndarray: The returned line signal
        """
        signal_type = self.__assess_line_sig()
        self.rectified_lines = self.__process_method[signal_type]()
        return self.rectified_lines

    def __attrs_post_init__(self):
        """
        Generate a basic dictionary to choose the right processing method
        :return:
        """
        self.__process_method = {
            LineSignalType.EXTRA_LINES: self.__truncate_extra_lines,
            LineSignalType.VALID_UNIDIR: self.__preprocess_unidir_lines,
            LineSignalType.VALID_BIDIR: self.__preprocess_bidir_lines,
            LineSignalType.SINGLE_PIXEL: self.__single_pixel,

        }


    def __assess_line_sig(self) -> LineSignalType:
        """
        Preliminary effort to detect the quality of the line signal.
        :return LineSignalType:
        """
        self.lines.sort()
        num_of_lines = len(self.lines)
        if num_of_lines > 1:
            if np.abs(1 - (num_of_lines / self.x_pixels)) > self.ALLOWED_THRESHOLD:  # line signal was too corrupt
                if num_of_lines >= self.x_pixels:  # corrupt, but due to some extra lines
                    return LineSignalType.EXTRA_LINES
                else:  # really too corrupt
                    raise ValueError

            else:  # lines signal is corrupt, but we can save it
                return LineSignalType.VALID_BIDIR if self.bidir else LineSignalType.VALID_UNIDIR

        else:  # single pixel frames, perhaps
            return LineSignalType.SINGLE_PIXEL

    def __truncate_extra_lines(self) -> np.ndarray:
        """
        Called when the line signal is both dirty and contains extra lines
        :return: np.ndarray
        """
        mean_diff = np.diff(self.lines).mean()
        return np.r_[
            self.lines[:self.x_pixels], np.array([self.lines[self.x_pixels - 1] + mean_diff], dtype=np.uint64)]

    def __single_pixel(self) -> np.ndarray:
        """
        Dumbed down case of a single-pixel frame
        :return:
        """
        return np.r_[self.lines, self.lines + self.end_time]

    def __preprocess_unidir_lines(self):
        """
        Obtain the needed parameters for the rectify_lines method
        :return: np.ndarray generated by the rectify_lines method
        """
        lines = pd.Series(self.lines)
        diffs = lines.diff()
        rel_idx = np.where(np.abs(diffs.pct_change(periods=1)) > self.ALLOWED_THRESHOLD)[0]
        rel_diff = np.diff(rel_idx) == 2
        if rel_diff.shape[0] > 0 and sum(rel_diff) > 0:
            for idx in np.where(np.concatenate((rel_diff, np.array([False]))))[0]:  # edge case
                rel_idx[idx + 1] -= 1
        mean_val = diffs.drop(rel_idx).mean()
        return self.__rectify_lines(lines=lines, diffs=diffs, mean_val=mean_val,
                                    rel_idx=rel_idx)

    def __preprocess_bidir_lines(self):
        """
        Obtain the needed parameters for the rectify_lines method
        :return: np.ndarray generated by the rectify_lines method
        """
        self.ALLOWED_THRESHOLD = 0.15  # x100 percent
        lines = pd.Series(self.lines)
        diff1 = lines.diff()
        diffs = lines.rolling(2).sum().diff()
        diffs[1] = diffs.mean()
        rel_idx = np.where(np.abs(diffs.pct_change(periods=1)) > self.ALLOWED_THRESHOLD)[0]
        if len(rel_idx) == 0:
            return self.__rectify_lines(lines=lines, diffs=diff1, mean_val=diff1.mean(),
                                        rel_idx=rel_idx)

        if len(rel_idx) % 2 != 0:  # every misplaced line should have a start and an end
            raise ValueError
        idx_to_throw = np.array([], dtype=np.int64)
        for idx, _ in enumerate(rel_idx[::2]):
            idx_to_throw = np.concatenate((idx_to_throw,
                                          np.arange(rel_idx[idx*2], rel_idx[idx*2+1] + 1)))

        final_rel_idx = rel_idx[::2]
        # Double jumps of rows missing
        rel_diff = np.diff(final_rel_idx) == 2
        if rel_diff.shape[0] > 0 and sum(rel_diff) > 0:
            for idx in np.where(np.concatenate((rel_diff, np.array([False]))))[0]:  # edge case
                final_rel_idx[idx + 1] -= 1

        # Check for multiline double spacing
        double = np.where(np.diff(rel_idx)[::2] == 4)[0]
        if len(double) > 0:
            for idx in double:
                indices_to_concat = [place for place in range(rel_idx[idx]+1, rel_idx[idx+1])]
                new_rel_idx = np.concatenate((rel_idx[:idx+1], rel_idx[idx+2:],
                                              np.array(indices_to_concat, dtype=np.int64)))
        else:
            new_rel_idx = sorted(rel_idx)[::2]
        final_rel_idx = np.unique(np.concatenate((final_rel_idx, new_rel_idx)))

        # Check for multiline triple spacing
        for idx in np.where(np.diff(rel_idx)[::2] == 5)[0]:
            indices_to_concat = [place for place in range(rel_idx[idx] + 1, rel_idx[idx + 1])]
            rel_idx = np.concatenate((rel_idx, np.array(indices_to_concat, dtype=np.int64)))
        else:
            rel_idx.sort()
            final_rel_idx = np.unique(np.concatenate((final_rel_idx, rel_idx)))

        mean_val = diff1.drop(idx_to_throw).mean()
        return self.__rectify_lines(lines=lines, diffs=diff1, mean_val=mean_val,
                                    rel_idx=final_rel_idx)

    def __rectify_lines(self, lines: pd.Series, diffs: pd.Series, mean_val: float,
                        rel_idx: np.ndarray) -> np.ndarray:
        """
        Rectify a semi-broken line signal.
        :return: Rectified lines
        """

        if rel_idx.shape[0] > 0.2 * self.x_pixels:  # too many missing lines
            raise ValueError

        if len(lines)-1 in rel_idx:  # last lines are missing
            needed_lines = 1 + (self.x_pixels+1 - (len(lines)+len(rel_idx)))
            if needed_lines < 1:
                raise ValueError
            lines = lines[:-1].append(pd.Series(np.linspace(start=lines.iloc[-2] + mean_val,
                                                            stop=lines.iloc[-2]+((needed_lines+1) * mean_val),
                                                            num=needed_lines, dtype=np.uint64)))
            rel_idx = rel_idx[rel_idx != len(diffs)-1]
        if np.abs((diffs.iloc[1] - mean_val) / mean_val) > self.ALLOWED_THRESHOLD:  # first line came late
            rel_idx = np.r_[rel_idx, 1]
        if len(rel_idx) > 0:  # straight-forward addition of line signals
            for val in rel_idx:
                missing_lines = int(np.around(diffs[val] / mean_val)) - 1
                lines = lines.append(pd.Series(np.linspace(start=lines.iloc[val - 1] + mean_val,
                                                           stop=lines.iloc[val], endpoint=False,
                                                           num=missing_lines, dtype=np.uint64)))

        if len(lines) != self.x_pixels:  # lines weren't recorded from the get-go
            if lines.iloc[0] == 0:  # we'll append at the end
                lines = lines.sort_values().reset_index(drop=True)
                needed_lines = np.abs(self.x_pixels - len(lines))
                lines = lines.append(pd.Series(np.linspace(start=lines.iloc[-1] + mean_val,
                                                           stop=lines.iloc[-1] + (needed_lines + 1) * mean_val,
                                                           endpoint=False,
                                                           num=needed_lines, dtype=np.uint64)))
            else:
                lines = lines.append(pd.Series(np.linspace(start=0, stop=lines.iloc[0], endpoint=False,
                                                           num=np.abs(self.x_pixels - len(lines)), dtype=np.uint64)))
        lines = lines.sort_values().reset_index(drop=True)
        lines = np.r_[lines.iloc[:self.x_pixels], np.array([lines.iloc[self.x_pixels - 1] + mean_val],
                                                           dtype=np.uint64)]
        return lines
